@@ -252,6 +252,63 @@ class AlgorithmNotSupported(RRSIGError):
         super(AlgorithmNotSupported, self).__init__(**kwargs)
         self.template_kwargs['algorithm_text'] = dns.dnssec.algorithm_to_text(self.template_kwargs['algorithm'])
 
+class AlgorithmValidationProhibited(RRSIGError):
+    '''
+    >>> e = AlgorithmValidationProhibited(algorithm=5)
+    >>> e.args
+    [5]
+    >>> e.description
+    'DNSSEC specification prohibits validation of RRSIGs with DNSSEC algorithm 5 (RSASHA1).'
+    '''
+
+    _abstract = False
+    code = 'ALGORITHM_VALIDATION_PROHIBITED'
+    description_template = "DNSSEC specification prohibits validation of RRSIGs with DNSSEC algorithm %(algorithm)d (%(algorithm_text)s)."
+    references = ['RFC 8624, Sec. 3.1']
+    required_params = ['algorithm']
+
+    def __init__(self, **kwargs):
+        super(AlgorithmValidationProhibited, self).__init__(**kwargs)
+        self.template_kwargs['algorithm_text'] = dns.dnssec.algorithm_to_text(self.template_kwargs['algorithm'])
+
+class AlgorithmProhibited(RRSIGError):
+    '''
+    >>> e = AlgorithmProhibited(algorithm=5)
+    >>> e.args
+    [5]
+    >>> e.description
+    'DNSSEC specification prohibits signing with DNSSEC algorithm 5 (RSASHA1).'
+    '''
+
+    _abstract = False
+    code = 'ALGORITHM_PROHIBITED'
+    description_template = "DNSSEC specification prohibits signing with DNSSEC algorithm %(algorithm)d (%(algorithm_text)s)."
+    references = ['RFC 8624, Sec. 3.1']
+    required_params = ['algorithm']
+
+    def __init__(self, **kwargs):
+        super(AlgorithmProhibited, self).__init__(**kwargs)
+        self.template_kwargs['algorithm_text'] = dns.dnssec.algorithm_to_text(self.template_kwargs['algorithm'])
+
+class AlgorithmNotRecommended(RRSIGError):
+    '''
+    >>> e = AlgorithmNotRecommended(algorithm=5)
+    >>> e.args
+    [5]
+    >>> e.description
+    'DNSSEC specification recommends not signing with DNSSEC algorithm 5 (RSASHA1).'
+    '''
+
+    _abstract = False
+    code = 'ALGORITHM_NOT_RECOMMENDED'
+    description_template = "DNSSEC specification recommends not signing with DNSSEC algorithm %(algorithm)d (%(algorithm_text)s)."
+    references = ['RFC 8624, Sec. 3.1']
+    required_params = ['algorithm']
+
+    def __init__(self, **kwargs):
+        super(AlgorithmNotRecommended, self).__init__(**kwargs)
+        self.template_kwargs['algorithm_text'] = dns.dnssec.algorithm_to_text(self.template_kwargs['algorithm'])
+
 class DNSKEYRevokedRRSIG(RRSIGError):
     '''
     >>> e = DNSKEYRevokedRRSIG()
@@ -448,6 +505,18 @@ class RRSIGBadLengthEd448(RRSIGBadLengthEdDSA):
 class DSError(DomainNameAnalysisError):
     pass
 
+class ReferralForDSQuery(DSError):
+    '''
+    >>> e = ReferralForDSQuery(parent='baz.')
+    >>> e.description
+    'The server(s) for the parent zone (baz.) responded with a referral instead of answering authoritatively for the DS RR type.'
+    '''
+    _abstract = False
+    code = 'REFERRAL_FOR_DS_QUERY'
+    description_template = 'The server(s) for the parent zone (%(parent)s) responded with a referral instead of answering authoritatively for the DS RR type.'
+    references = ['RFC 4034, Sec. 5']
+    required_params = ['parent']
+
 class DSDigestAlgorithmIgnored(DSError):
     '''
     >>> e = DSDigestAlgorithmIgnored(algorithm=1, new_algorithm=2)
@@ -465,7 +534,24 @@ class DSDigestAlgorithmIgnored(DSError):
         self.template_kwargs['algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['algorithm'], str(self.template_kwargs['algorithm']))
         self.template_kwargs['new_algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['new_algorithm'], str(self.template_kwargs['algorithm']))
 
-class DSDigestError(DomainNameAnalysisError):
+class DSDigestAlgorithmMaybeIgnored(DSError):
+    '''
+    >>> e = DSDigestAlgorithmMaybeIgnored(algorithm=1, new_algorithm=2)
+    >>> e.description
+    'In the spirit of RFC 4509, DS records with digest type 1 (SHA-1) might be ignored when DS records with digest type 2 (SHA-256) exist in the same RRset.'
+    '''
+    _abstract = False
+    code = 'DS_DIGEST_ALGORITHM_MAYBE_IGNORED'
+    description_template = "In the spirit of RFC 4509, DS records with digest type %(algorithm)d (%(algorithm_text)s) might be ignored when DS records with digest type %(new_algorithm)d (%(new_algorithm_text)s) exist in the same RRset."
+    references = ['RFC 4509, Sec. 3']
+    required_params = ['algorithm', 'new_algorithm']
+
+    def __init__(self, **kwargs):
+        super(DSDigestAlgorithmMaybeIgnored, self).__init__(**kwargs)
+        self.template_kwargs['algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['algorithm'], str(self.template_kwargs['algorithm']))
+        self.template_kwargs['new_algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['new_algorithm'], str(self.template_kwargs['algorithm']))
+
+class DSDigestError(DSError):
     pass
 
 class DigestAlgorithmNotSupported(DSDigestError):
@@ -485,6 +571,57 @@ class DigestAlgorithmNotSupported(DSDigestError):
         super(DigestAlgorithmNotSupported, self).__init__(**kwargs)
         self.template_kwargs['algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['algorithm'], self.template_kwargs['algorithm'])
 
+class DigestAlgorithmValidationProhibited(DSDigestError):
+    '''
+    >>> e = DigestAlgorithmValidationProhibited(algorithm=5)
+    >>> e.description
+    'DNSSEC specification prohibits validation of DS records that use digest algorithm 5 (5).'
+    '''
+
+    _abstract = False
+    code = 'DIGEST_ALGORITHM_VALIDATION_PROHIBITED'
+    description_template = "DNSSEC specification prohibits validation of DS records that use digest algorithm %(algorithm)d (%(algorithm_text)s)."
+    references = ['RFC 8624, Sec. 3.2']
+    required_params = ['algorithm']
+
+    def __init__(self, **kwargs):
+        super(DigestAlgorithmValidationProhibited, self).__init__(**kwargs)
+        self.template_kwargs['algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['algorithm'], self.template_kwargs['algorithm'])
+
+class DigestAlgorithmProhibited(DSDigestError):
+    '''
+    >>> e = DigestAlgorithmProhibited(algorithm=5)
+    >>> e.description
+    'DNSSEC specification prohibits signing with DS records that use digest algorithm 5 (5).'
+    '''
+
+    _abstract = False
+    code = 'DIGEST_ALGORITHM_PROHIBITED'
+    description_template = "DNSSEC specification prohibits signing with DS records that use digest algorithm %(algorithm)d (%(algorithm_text)s)."
+    references = ['RFC 8624, Sec. 3.2']
+    required_params = ['algorithm']
+
+    def __init__(self, **kwargs):
+        super(DigestAlgorithmProhibited, self).__init__(**kwargs)
+        self.template_kwargs['algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['algorithm'], self.template_kwargs['algorithm'])
+
+class DigestAlgorithmNotRecommended(DSDigestError):
+    '''
+    >>> e = DigestAlgorithmNotRecommended(algorithm=5)
+    >>> e.description
+    'DNSSEC specification recommends not signing with DS records that use digest algorithm 5 (5).'
+    '''
+
+    _abstract = False
+    code = 'DIGEST_ALGORITHM_NOT_RECOMMENDED'
+    description_template = "DNSSEC specification recommends not signing with DS records that use digest algorithm %(algorithm)d (%(algorithm_text)s)."
+    references = ['RFC 8624, Sec. 3.2']
+    required_params = ['algorithm']
+
+    def __init__(self, **kwargs):
+        super(DigestAlgorithmNotRecommended, self).__init__(**kwargs)
+        self.template_kwargs['algorithm_text'] = fmt.DS_DIGEST_TYPES.get(self.template_kwargs['algorithm'], self.template_kwargs['algorithm'])
+
 class DNSKEYRevokedDS(DSDigestError):
     '''
     >>> e = DNSKEYRevokedDS()
@@ -830,6 +967,66 @@ class WildcardCoveredNODATANSEC3(WildcardCoveredNODATA):
     references = ['RFC 5155, Sec. 8.7']
     nsec_type = 'NSEC3'
 
+class ExistingNSECError(NSECError):
+    required_params = ['queries']
+
+    def __init__(self, **kwargs):
+        super(ExistingNSECError, self).__init__(**kwargs)
+        queries_text = ['%s/%s' % (name, rdtype) for name, rdtype in self.template_kwargs['queries']]
+        self.template_kwargs['queries_text'] = ', '.join(queries_text)
+
+class ExistingCovered(ExistingNSECError):
+    description_template = 'The following queries resulted in an answer response, even though the %(nsec_type)s records indicate that the queried names don\'t exist: %(queries_text)s'
+    code = 'EXISTING_NAME_COVERED'
+
+class ExistingCoveredNSEC(ExistingCovered):
+    '''
+    >>> e = ExistingCoveredNSEC(queries=[('www.foo.baz.', 'A'), ('www1.foo.baz.', 'TXT')])
+    >>> e.description
+    "The following queries resulted in an answer response, even though the NSEC records indicate that the queried names don't exist: www.foo.baz./A, www1.foo.baz./TXT"
+    '''
+
+    _abstract = False
+    references = ['RFC 4035, Sec. 3.1.3.2']
+    nsec_type = 'NSEC'
+
+class ExistingCoveredNSEC3(ExistingCovered):
+    '''
+    >>> e = ExistingCoveredNSEC3(queries=[('www.foo.baz.', 'A'), ('www1.foo.baz.', 'TXT')])
+    >>> e.description
+    "The following queries resulted in an answer response, even though the NSEC3 records indicate that the queried names don't exist: www.foo.baz./A, www1.foo.baz./TXT"
+    '''
+
+    _abstract = False
+    references = ['RFC 5155, Sec. 8.4']
+    nsec_type = 'NSEC3'
+
+class ExistingTypeNotInBitmap(ExistingNSECError):
+    description_template = 'The following queries resulted in an answer response, even though the bitmap in the %(nsec_type)s RR indicates that the queried records don\'t exist: %(queries_text)s'
+    code = 'EXISTING_TYPE_NOT_IN_BITMAP'
+
+class ExistingTypeNotInBitmapNSEC(ExistingTypeNotInBitmap):
+    '''
+    >>> e = ExistingTypeNotInBitmapNSEC(queries=[('www.foo.baz.', 'A'), ('www.foo.baz.', 'TXT')])
+    >>> e.description
+    "The following queries resulted in an answer response, even though the bitmap in the NSEC RR indicates that the queried records don't exist: www.foo.baz./A, www.foo.baz./TXT"
+    '''
+
+    _abstract = False
+    references = ['RFC 4035, Sec. 3.1.3.1']
+    nsec_type = 'NSEC'
+
+class ExistingTypeNotInBitmapNSEC3(ExistingTypeNotInBitmap):
+    '''
+    >>> e = ExistingTypeNotInBitmapNSEC3(queries=[('www.foo.baz.', 'A'), ('www.foo.baz.', 'TXT')])
+    >>> e.description
+    "The following queries resulted in an answer response, even though the bitmap in the NSEC3 RR indicates that the queried records don't exist: www.foo.baz./A, www.foo.baz./TXT"
+    '''
+
+    _abstract = False
+    references = ['RFC 5155, Sec. 8.5']
+    nsec_type = 'NSEC3'
+
 class SnameCoveredNODATANSEC(NSECError):
     '''
     >>> e = SnameCoveredNODATANSEC(sname='foo.baz.')
@@ -1257,15 +1454,53 @@ class EDNSUndefinedFlagsSet(EDNSError):
     '''
     >>> e = EDNSUndefinedFlagsSet(flags=0x80)
     >>> e.description
-    'The server set EDNS flags that are undefined: 128.'
+    'The server set EDNS flags that are undefined: 0x80.'
     '''
 
     _abstract = False
     code = 'EDNS_UNDEFINED_FLAGS_SET'
-    description_template = 'The server set EDNS flags that are undefined: %(flags)d.'
+    description_template = 'The server set EDNS flags that are undefined: %(flags_text)s.'
     references = ['RFC 6891, Sec. 6.1.4']
     required_params = ['flags']
 
+    def __init__(self, **kwargs):
+        super(EDNSUndefinedFlagsSet, self).__init__(**kwargs)
+        self.template_kwargs['flags_text'] = '0x%x' % (self.template_kwargs['flags'])
+
+class DNSSECDowngrade(EDNSError):
+    description_template = "DNSSEC was effectively downgraded because %(response_error_description)s with %(precondition)s."
+    required_params = ['response_error']
+    precondition = None
+
+    def __init__(self, *args, **kwargs):
+        super(DNSSECDowngrade, self).__init__(**kwargs)
+        self.template_kwargs['response_error_description'] = self.template_kwargs['response_error'].description[0].lower() + self.template_kwargs['response_error'].description[1:-1]
+        self.template_kwargs['precondition'] = self.precondition
+
+class DNSSECDowngradeDOBitCleared(DNSSECDowngrade):
+    '''
+    >>> e = DNSSECDowngradeDOBitCleared(response_error=Timeout(tcp=False, attempts=3))
+    >>> e.description
+    'DNSSEC was effectively downgraded because no response was received from the server over UDP (tried 3 times) with the DO bit set.'
+    '''
+
+    _abstract = False
+    code = 'DNSSEC_DOWNGRADE_DO_CLEARED'
+    precondition = 'the DO bit set'
+    references = ['RFC 4035, Sec. 3.2.1']
+
+class DNSSECDowngradeEDNSDisabled(DNSSECDowngrade):
+    '''
+    >>> e = DNSSECDowngradeEDNSDisabled(response_error=Timeout(tcp=False, attempts=3), query_specific=False)
+    >>> e.description
+    'DNSSEC was effectively downgraded because no response was received from the server over UDP (tried 3 times) with EDNS enabled.'
+    '''
+
+    _abstract = False
+    code = 'DNSSEC_DOWNGRADE_EDNS_DISABLED'
+    precondition = 'EDNS enabled'
+    references = ['RFC 6891, Sec. 7', 'RFC 2671, Sec. 5.3']
+
 class DNSCookieError(ResponseError):
     pass
 
@@ -1811,6 +2046,31 @@ class NoAddressForNSName(NSNameError):
     code = 'NO_ADDRESS_FOR_NS_NAME'
     description_template = "The following NS name(s) did not resolve to address(es): %(names_text)s"
 
+class PrivateAddressNS(NSNameError):
+    pass
+
+class NSNameResolvesToPrivateIP(PrivateAddressNS):
+    '''
+    >>> e = NSNameResolvesToPrivateIP(names=('ns1.foo.baz.',))
+    >>> e.description
+    'The following NS name(s) resolved to IP address(es) in private IP address space: ns1.foo.baz.'
+    '''
+
+    _abstract = False
+    code = 'NS_NAME_PRIVATE_IP'
+    description_template = "The following NS name(s) resolved to IP address(es) in private IP address space: %(names_text)s"
+
+class GlueReferencesPrivateIP(PrivateAddressNS):
+    '''
+    >>> e = GlueReferencesPrivateIP(names=('ns1.foo.baz.',))
+    >>> e.description
+    'Glue for the following NS name(s) referenced IP address(es) in private IP address space: ns1.foo.baz.'
+    '''
+
+    _abstract = False
+    code = 'GLUE_PRIVATE_IP'
+    description_template = "Glue for the following NS name(s) referenced IP address(es) in private IP address space: %(names_text)s"
+
 class GlueMismatchError(DelegationError):
     '''
     >>> e = GlueMismatchError(name='ns1.foo.baz.', glue_addresses=('192.0.2.1',), auth_addresses=('192.0.2.2',))
@@ -2011,6 +2271,18 @@ class DNSKEYNotAtZoneApex(DNSKEYError):
 class DNSKEYBadLength(DNSKEYError):
     pass
 
+class DNSKEYZeroLength(DNSKEYBadLength):
+    '''
+    >>> e = DNSKEYZeroLength()
+    >>> e.description
+    'The length of the key is 0 bits.'
+    '''
+    _abstract = False
+    description_template = 'The length of the key is 0 bits.'
+    code = 'DNSKEY_ZERO_LENGTH'
+    references = []
+    required_params = []
+
 class DNSKEYBadLengthGOST(DNSKEYBadLength):
     '''
     >>> e = DNSKEYBadLengthGOST(length=500)
@@ -2137,3 +2409,14 @@ class CNAMEWithOtherData(ZoneDataError):
     code = 'CNAME_WITH_OTHER_DATA'
     required_params = ['name']
     references = ['RFC 2181, Sec. 10.1']
+
+class CNAMELoop(ZoneDataError):
+    '''
+    >>> e = CNAMELoop()
+    >>> e.description
+    'This record results in a CNAME loop.'
+    '''
+    _abstract = False
+    description_template = "This record results in a CNAME loop."
+    code = 'CNAME_LOOP'
+    references = ['RFC 1034, Sec. 3.6.2']
