@@ -166,6 +166,18 @@ RRSIG_SIG_LENGTH_ERRORS = {
         14: Errors.RRSIGBadLengthECDSA384, 15: Errors.RRSIGBadLengthEd25519,
         16: Errors.RRSIGBadLengthEd448,
 }
+DS_DIGEST_ALGS_STRONGER_THAN_SHA1 = (2, 4)
+DS_DIGEST_ALGS_IGNORING_SHA1 = (2,)
+
+# RFC 8624 Section 3.1
+DNSKEY_ALGS_NOT_RECOMMENDED = (5, 7, 10)
+DNSKEY_ALGS_PROHIBITED = (1, 3, 6, 12)
+DNSKEY_ALGS_VALIDATION_PROHIBITED = (1, 3, 6)
+
+# RFC 8624 Section 3.2
+DS_DIGEST_ALGS_NOT_RECOMMENDED = ()
+DS_DIGEST_ALGS_PROHIBITED = (0, 1, 3)
+DS_DIGEST_ALGS_VALIDATION_PROHIBITED = ()
 
 class RRSIGStatus(object):
     def __init__(self, rrset, rrsig, dnskey, zone_name, reference_ts, supported_algs):
@@ -184,13 +196,44 @@ class RRSIGStatus(object):
 
         self.validation_status = RRSIG_STATUS_VALID
         if self.signature_valid is None or self.dnskey.rdata.algorithm not in supported_algs:
+            # Either we can't validate the cryptographic signature, or we are
+            # explicitly directed to ignore the algorithm.
             if self.dnskey is None:
+                # In this case, there is no corresponding DNSKEY, so we make
+                # the status "INDETERMINATE".
                 if self.validation_status == RRSIG_STATUS_VALID:
                     self.validation_status = RRSIG_STATUS_INDETERMINATE_NO_DNSKEY
+
             else:
-                if self.validation_status == RRSIG_STATUS_VALID:
-                    self.validation_status = RRSIG_STATUS_INDETERMINATE_UNKNOWN_ALGORITHM
-                self.warnings.append(Errors.AlgorithmNotSupported(algorithm=self.rrsig.algorithm))
+                # If there is a DNSKEY, then we look at *why* we are ignoring
+                # the cryptographic signature.
+                if self.dnskey.rdata.algorithm in DNSKEY_ALGS_VALIDATION_PROHIBITED:
+                    # In this case, specification dictates that the algorithm
+                    # MUST NOT be validated, so we mark it as ignored.
+                    if self.validation_status == RRSIG_STATUS_VALID:
+                        self.validation_status = RRSIG_STATUS_ALGORITHM_IGNORED
+                else:
+                    # In this case, we can't validate this particular
+                    # algorithm, either because the code doesn't support it,
+                    # or because we have been explicitly directed to ignore it.
+                    # In either case, mark it as "UNKNOWN", and warn that it is
+                    # not supported.
+                    if self.validation_status == RRSIG_STATUS_VALID:
+                        self.validation_status = RRSIG_STATUS_INDETERMINATE_UNKNOWN_ALGORITHM
+                    self.warnings.append(Errors.AlgorithmNotSupported(algorithm=self.rrsig.algorithm))
+
+        # Independent of whether or not we considered the cryptographic
+        # validation, issue a warning if we are using an algorithm for which
+        # validation or signing has been prohibited.
+        #
+        # Signing is prohibited
+        if self.dnskey.rdata.algorithm in DNSKEY_ALGS_VALIDATION_PROHIBITED:
+            self.warnings.append(Errors.AlgorithmValidationProhibited(algorithm=self.rrsig.algorithm))
+        # Validation is prohibited or, at least, not recommended
+        if self.dnskey.rdata.algorithm in DNSKEY_ALGS_PROHIBITED:
+            self.warnings.append(Errors.AlgorithmProhibited(algorithm=self.rrsig.algorithm))
+        elif self.dnskey.rdata.algorithm in DNSKEY_ALGS_NOT_RECOMMENDED:
+            self.warnings.append(Errors.AlgorithmNotRecommended(algorithm=self.rrsig.algorithm))
 
         if self.rrset.ttl_cmp:
             if self.rrset.rrset.ttl != self.rrset.rrsig_info[self.rrsig].ttl:
@@ -251,7 +294,7 @@ class RRSIGStatus(object):
     def __str__(self):
         return 'RRSIG covering %s/%s' % (fmt.humanize_name(self.rrset.rrset.name), dns.rdatatype.to_text(self.rrset.rrset.rdtype))
 
-    def serialize(self, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         d = OrderedDict()
 
         erroneous_status = self.validation_status not in (RRSIG_STATUS_VALID, RRSIG_STATUS_INDETERMINATE_NO_DNSKEY, RRSIG_STATUS_INDETERMINATE_UNKNOWN_ALGORITHM)
@@ -304,11 +347,25 @@ class RRSIGStatus(object):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.rrset.rrsig_info[self.rrsig].servers_clients:
                 for response in self.rrset.rrsig_info[self.rrsig].servers_clients[(server, client)]:
                     if response is not None:
                         tags.add(response.effective_query_tag())
+                        nsid = response.nsid_val()
+                        if nsid is not None:
+                            nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -335,13 +392,43 @@ class DSStatus(object):
 
         self.validation_status = DS_STATUS_VALID
         if self.digest_valid is None or self.ds.digest_type not in supported_digest_algs:
+            # Either we cannot reproduce a digest with this type, or we are
+            # explicitly directed to ignore the digest type.
             if self.dnskey is None:
+                # In this case, there is no corresponding DNSKEY, so we make
+                # the status "INDETERMINATE".
                 if self.validation_status == DS_STATUS_VALID:
                     self.validation_status = DS_STATUS_INDETERMINATE_NO_DNSKEY
             else:
-                if self.validation_status == DS_STATUS_VALID:
-                    self.validation_status = DS_STATUS_INDETERMINATE_UNKNOWN_ALGORITHM
-                self.warnings.append(Errors.DigestAlgorithmNotSupported(algorithm=ds.digest_type))
+                # If there is a DNSKEY, then we look at *why* we are ignoring
+                # the digest of the DNSKEY.
+                if self.ds.digest_type in DS_DIGEST_ALGS_VALIDATION_PROHIBITED:
+                    # In this case, specification dictates that the algorithm
+                    # MUST NOT be validated, so we mark it as ignored.
+                    if self.validation_status == DS_STATUS_VALID:
+                        self.validation_status = DS_STATUS_ALGORITHM_IGNORED
+                else:
+                    # In this case, we can't validate this particular
+                    # digest type, either because the code doesn't support it,
+                    # or because we have been explicitly directed to ignore it.
+                    # In either case, mark it as "UNKNOWN", and warn that it is
+                    # not supported.
+                    if self.validation_status == DS_STATUS_VALID:
+                        self.validation_status = DS_STATUS_INDETERMINATE_UNKNOWN_ALGORITHM
+                    self.warnings.append(Errors.DigestAlgorithmNotSupported(algorithm=self.ds.digest_type))
+
+        # Independent of whether or not we considered the digest for
+        # validation, issue a warning if we are using a digest type for which
+        # validation or signing has been prohibited.
+        #
+        # Signing is prohibited
+        if self.ds.digest_type in DS_DIGEST_ALGS_VALIDATION_PROHIBITED:
+            self.warnings.append(Errors.DigestAlgorithmValidationProhibited(algorithm=self.ds.digest_type))
+        # Validation is prohibited or, at least, not recommended
+        if self.ds.digest_type in DS_DIGEST_ALGS_PROHIBITED:
+            self.warnings.append(Errors.DigestAlgorithmProhibited(algorithm=self.ds.digest_type))
+        elif self.ds.digest_type in DS_DIGEST_ALGS_NOT_RECOMMENDED:
+            self.warnings.append(Errors.DigestAlgorithmNotRecommended(algorithm=self.ds.digest_type))
 
         if self.dnskey is not None and \
                 self.dnskey.rdata.flags & fmt.DNSKEY_FLAGS['revoke']:
@@ -362,35 +449,33 @@ class DSStatus(object):
 
         # RFC 4509
         if self.ds.digest_type == 1:
-            digest_algs = set()
-            my_digest_algs = set()
-            my_digest_algs = {}
+            stronger_algs_all_ds = set()
+            # Cycle through all other DS records in the DS RRset, and
+            # create a list of digest types that are stronger than SHA1
+            # and are being used by DS records across the *entire* DS.
             for ds_rdata in self.ds_meta.rrset:
-                digest_algs.add(ds_rdata.digest_type)
-                if (ds_rdata.algorithm, ds_rdata.key_tag) == (self.ds.algorithm, self.ds.key_tag):
-                    # Here we produce a status of the DS with algorithm 2 with
-                    # respect to the DNSKEY for comparison with the current DS
-                    # with algorithm 1.  It is possible that the DS with the
-                    # different digest type matches a different DNSKEY than the
-                    # present DNSKEY.  If it does, there will be a warning, but
-                    # it should be both rare and innocuous.
-                    if ds_rdata.digest_type == self.ds.digest_type:
-                        continue
-                    elif ds_rdata.digest_type not in my_digest_algs or \
-                            my_digest_algs[ds_rdata.digest_type].validation_status != DS_STATUS_VALID:
-                        my_digest_algs[ds_rdata.digest_type] = \
-                                DSStatus(ds_rdata, self.ds_meta, self.dnskey, supported_digest_algs)
-            if 2 in supported_digest_algs and 2 in digest_algs and \
-                    (2 not in my_digest_algs or my_digest_algs[2].validation_status not in \
-                    (DS_STATUS_VALID, DS_STATUS_INDETERMINATE_NO_DNSKEY, DS_STATUS_INDETERMINATE_UNKNOWN_ALGORITHM, DS_STATUS_INDETERMINATE_MATCH_PRE_REVOKE)):
-                self.warnings.append(Errors.DSDigestAlgorithmIgnored(algorithm=1, new_algorithm=2))
-                if self.validation_status == DS_STATUS_VALID:
-                    self.validation_status = DS_STATUS_ALGORITHM_IGNORED
+                if ds_rdata.digest_type in DS_DIGEST_ALGS_STRONGER_THAN_SHA1:
+                    stronger_algs_all_ds.add(ds_rdata.digest_type)
+
+            # Consider only digest types that we actually support
+            stronger_algs_all_ds.intersection_update(supported_digest_algs)
+
+            if stronger_algs_all_ds:
+                # If there are DS records in the DS RRset with digest type
+                # stronger than SHA1, then this one MUST be ignored by
+                # validators (RFC 4509).
+                for digest_alg in stronger_algs_all_ds:
+                    if digest_alg in DS_DIGEST_ALGS_IGNORING_SHA1:
+                        if self.validation_status == DS_STATUS_VALID:
+                            self.validation_status = DS_STATUS_ALGORITHM_IGNORED
+                        self.warnings.append(Errors.DSDigestAlgorithmIgnored(algorithm=1, new_algorithm=digest_alg))
+                    else:
+                        self.warnings.append(Errors.DSDigestAlgorithmMaybeIgnored(algorithm=1, new_algorithm=digest_alg))
 
     def __str__(self):
         return '%s record(s) corresponding to DNSKEY for %s (algorithm %d (%s), key tag %d)' % (dns.rdatatype.to_text(self.ds_meta.rrset.rdtype), fmt.humanize_name(self.ds_meta.rrset.name), self.ds.algorithm, fmt.DNSKEY_ALGORITHMS.get(self.ds.algorithm, self.ds.algorithm), self.ds.key_tag)
 
-    def serialize(self, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=True):
         d = OrderedDict()
 
         erroneous_status = self.validation_status not in (DS_STATUS_VALID, DS_STATUS_INDETERMINATE_NO_DNSKEY, DS_STATUS_INDETERMINATE_UNKNOWN_ALGORITHM)
@@ -435,11 +520,25 @@ class DSStatus(object):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.ds_meta.servers_clients:
                 for response in self.ds_meta.servers_clients[(server, client)]:
                     if response is not None:
                         tags.add(response.effective_query_tag())
+                        nsid = response.nsid_val()
+                        if nsid is not None:
+                            nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -480,6 +579,8 @@ class NSECStatusNXDOMAIN(NSECStatus):
 
         self.nsec_names_covering_qname = {}
         covering_names = nsec_set_info.nsec_covering_name(self.qname)
+        self.opt_out = None
+
         if covering_names:
             self.nsec_names_covering_qname[self.qname] = covering_names
 
@@ -534,7 +635,7 @@ class NSECStatusNXDOMAIN(NSECStatus):
     def __str__(self):
         return 'NSEC record(s) proving the non-existence (NXDOMAIN) of %s' % (fmt.humanize_name(self.qname))
 
-    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         d = OrderedDict()
 
         nsec_list = []
@@ -597,11 +698,25 @@ class NSECStatusNXDOMAIN(NSECStatus):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.nsec_set_info.servers_clients:
                 for response in self.nsec_set_info.servers_clients[(server, client)]:
                     if response is not None:
                         tags.add(response.effective_query_tag())
+                        nsid = response.nsid_val()
+                        if nsid is not None:
+                            nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -660,8 +775,8 @@ class NSECStatusWildcard(NSECStatusNXDOMAIN):
         else:
             self.nsec_set_info = nsec_set_info.project(*list(nsec_set_info.rrsets))
 
-    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
-        d = super(NSECStatusWildcard, self).serialize(rrset_info_serializer, consolidate_clients=consolidate_clients, loglevel=loglevel, html_format=html_format)
+    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
+        d = super(NSECStatusWildcard, self).serialize(rrset_info_serializer, consolidate_clients=consolidate_clients, loglevel=loglevel, html_format=html_format, map_ip_to_ns_name=map_ip_to_ns_name)
         try:
             del d['wildcard']
         except KeyError:
@@ -669,7 +784,7 @@ class NSECStatusWildcard(NSECStatusNXDOMAIN):
         return d
 
 class NSECStatusNODATA(NSECStatus):
-    def __init__(self, qname, rdtype, origin, is_zone, nsec_set_info):
+    def __init__(self, qname, rdtype, origin, is_zone, nsec_set_info, sname_must_match=False):
         self.qname = qname
         self.rdtype = rdtype
         self.origin = origin
@@ -693,13 +808,14 @@ class NSECStatusNODATA(NSECStatus):
             self.has_ds = False
             self.has_soa = False
 
-            # If no NSEC exists for the name itself, then look for an NSEC with
-            # an (empty non-terminal) ancestor
-            for nsec_name in nsec_set_info.rrsets:
-                next_name = nsec_set_info.rrsets[nsec_name].rrset[0].next
-                if next_name.is_subdomain(self.qname) and next_name != self.qname:
-                    self.nsec_for_qname = nsec_set_info.rrsets[nsec_name]
-                    break
+            if not sname_must_match:
+                # If no NSEC exists for the name itself, then look for an NSEC with
+                # an (empty non-terminal) ancestor
+                for nsec_name in nsec_set_info.rrsets:
+                    next_name = nsec_set_info.rrsets[nsec_name].rrset[0].next
+                    if next_name.is_subdomain(self.qname) and next_name != self.qname:
+                        self.nsec_for_qname = nsec_set_info.rrsets[nsec_name]
+                        break
 
         self.nsec_names_covering_qname = {}
         covering_names = nsec_set_info.nsec_covering_name(self.qname)
@@ -724,6 +840,8 @@ class NSECStatusNODATA(NSECStatus):
         if covering_names:
             self.nsec_names_covering_origin[self.origin] = covering_names
 
+        self.opt_out = None
+
         self._set_validation_status(nsec_set_info)
 
     def __str__(self):
@@ -785,7 +903,7 @@ class NSECStatusNODATA(NSECStatus):
         else:
             self.nsec_set_info = nsec_set_info.project(*list(nsec_set_info.rrsets))
 
-    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         d = OrderedDict()
 
         nsec_list = []
@@ -845,11 +963,25 @@ class NSECStatusNODATA(NSECStatus):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.nsec_set_info.servers_clients:
                 for response in self.nsec_set_info.servers_clients[(server, client)]:
                     if response is not None:
                         tags.add(response.effective_query_tag())
+                        nsid = response.nsid_val()
+                        if nsid is not None:
+                            nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -998,7 +1130,7 @@ class NSEC3StatusNXDOMAIN(NSEC3Status):
         for name in self.nsec_set_info.invalid_nsec3_hash:
             self.errors.append(Errors.InvalidNSEC3Hash(name=fmt.format_nsec3_name(name), nsec3_hash=lb2s(base32.b32encode(self.nsec_set_info.rrsets[name].rrset[0].next))))
 
-    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         d = OrderedDict()
 
         nsec3_list = []
@@ -1095,11 +1227,25 @@ class NSEC3StatusNXDOMAIN(NSEC3Status):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.nsec_set_info.servers_clients:
                 for response in self.nsec_set_info.servers_clients[(server, client)]:
                     if response is not None:
                         tags.add(response.effective_query_tag())
+                        nsid = response.nsid_val()
+                        if nsid is not None:
+                            nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -1166,8 +1312,8 @@ class NSEC3StatusWildcard(NSEC3StatusNXDOMAIN):
         for name in self.nsec_set_info.invalid_nsec3_hash:
             self.errors.append(Errors.InvalidNSEC3Hash(name=fmt.format_nsec3_name(name), nsec3_hash=lb2s(base32.b32encode(self.nsec_set_info.rrsets[name].rrset[0].next))))
 
-    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
-        d = super(NSEC3StatusWildcard, self).serialize(rrset_info_serializer, consolidate_clients=consolidate_clients, loglevel=loglevel, html_format=html_format)
+    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
+        d = super(NSEC3StatusWildcard, self).serialize(rrset_info_serializer, consolidate_clients=consolidate_clients, loglevel=loglevel, html_format=html_format, map_ip_to_ns_name=map_ip_to_ns_name)
         try:
             del d['wildcard']
         except KeyError:
@@ -1302,7 +1448,7 @@ class NSEC3StatusNODATA(NSEC3Status):
                     self.errors.append(invalid_alg_err)
             if self.wildcard_has_rdtype:
                 self.validation_status = NSEC_STATUS_INVALID
-                self.errors.append(Errors.StypeInBitmapWildcardNODATANSEC3(sname=fmt.humanize_name(self.wildcard_name), stype=dns.rdatatype.to_text(self.rdtype)))
+                self.errors.append(Errors.StypeInBitmapWildcardNODATANSEC3(sname=fmt.humanize_name(self.get_wildcard()), stype=dns.rdatatype.to_text(self.rdtype)))
         elif self.nsec_names_covering_qname:
             if not self.opt_out:
                 self.validation_status = NSEC_STATUS_INVALID
@@ -1349,7 +1495,7 @@ class NSEC3StatusNODATA(NSEC3Status):
         for name in self.nsec_set_info.invalid_nsec3_hash:
             self.errors.append(Errors.InvalidNSEC3Hash(name=fmt.format_nsec3_name(name), nsec3_hash=lb2s(base32.b32encode(self.nsec_set_info.rrsets[name].rrset[0].next))))
 
-    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         d = OrderedDict()
 
         nsec3_list = []
@@ -1445,11 +1591,25 @@ class NSEC3StatusNODATA(NSEC3Status):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.nsec_set_info.servers_clients:
                 for response in self.nsec_set_info.servers_clients[(server, client)]:
                     if response is not None:
                         tags.add(response.effective_query_tag())
+                        nsid = response.nsid_val()
+                        if nsid is not None:
+                            nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -1485,7 +1645,7 @@ class CNAMEFromDNAMEStatus(object):
     def __str__(self):
         return 'CNAME synthesis for %s from %s/%s' % (fmt.humanize_name(self.synthesized_cname.rrset.name), fmt.humanize_name(self.synthesized_cname.dname_info.rrset.name), dns.rdatatype.to_text(self.synthesized_cname.dname_info.rrset.rdtype))
 
-    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, rrset_info_serializer=None, consolidate_clients=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         values = []
         d = OrderedDict()
 
@@ -1531,11 +1691,25 @@ class CNAMEFromDNAMEStatus(object):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.synthesized_cname.dname_info.servers_clients:
                 for response in self.synthesized_cname.dname_info.servers_clients[(server, client)]:
                     if response is not None:
                         tags.add(response.effective_query_tag())
+                        nsid = response.nsid_val()
+                        if nsid is not None:
+                            nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
