@@ -52,7 +52,7 @@ try:
 except ImportError:
     from cgi import escape
 
-import dns.flags, dns.message, dns.rcode, dns.rdataclass, dns.rdatatype, dns.rrset
+import dns.edns, dns.flags, dns.message, dns.rcode, dns.rdataclass, dns.rdatatype, dns.rrset
 
 from . import base32
 from . import crypto
@@ -209,6 +209,24 @@ class DNSResponse:
             s += '_0x20'
         return s
 
+    def nsid_val(self):
+        if self.message is None:
+            return None
+
+        if self.message.edns < 0:
+            return None
+
+        try:
+            nsid_opt = [o for o in self.message.options if o.otype == dns.edns.NSID][0]
+        except IndexError:
+            return None
+
+        try:
+            nsid_val = nsid_opt.data.decode('ascii')
+        except UnicodeDecodeError:
+            nsid_val = '0x' + lb2s(binascii.hexlify(nsid_opt.data))
+        return nsid_val
+
     def request_cookie_tag(self):
         from . import query as Q
 
@@ -860,8 +878,8 @@ class DNSKEYMeta(DNSResponseComponent):
 
             offset = 1
             if e_len == 0:
-                b1 = map_func(rdata.key[1])
-                b2 = map_func(rdata.key[2])
+                b1 = map_func(key_str, 1)
+                b2 = map_func(key_str, 2)
                 e_len = (b1 << 8) | b2
                 offset = 3
 
@@ -914,7 +932,7 @@ class DNSKEYMeta(DNSResponseComponent):
 
         return name_wire + rdata_wire + self.rdata.key
 
-    def serialize(self, consolidate_clients=True, show_servers=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, consolidate_clients=True, show_servers=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         from .analysis import status as Status
 
         show_id = loglevel <= logging.INFO or \
@@ -962,10 +980,24 @@ class DNSKEYMeta(DNSResponseComponent):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.servers_clients:
                 for response in self.servers_clients[(server,client)]:
                     tags.add(response.effective_query_tag())
+                    nsid = response.nsid_val()
+                    if nsid is not None:
+                        nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -1068,7 +1100,7 @@ class RRsetInfo(DNSResponseComponent):
         return self.insert_into_list(synthesized_cname_info, self.cname_info_from_dname, server, client, response)
 
     def is_wildcard(self, rrsig):
-        if self.rrset.name[0] == '*':
+        if self.rrset.name[0] == b'*':
             return False
         return len(self.rrset.name) - 1 > rrsig.labels
 
@@ -1100,7 +1132,7 @@ class RRsetInfo(DNSResponseComponent):
 
         return rrsig_canonicalized_wire + rrset_canonicalized_wire
 
-    def serialize(self, consolidate_clients=True, show_servers=True, loglevel=logging.DEBUG, html_format=False):
+    def serialize(self, consolidate_clients=True, show_servers=True, loglevel=logging.DEBUG, html_format=False, map_ip_to_ns_name=None):
         d = OrderedDict()
 
         if html_format:
@@ -1135,10 +1167,24 @@ class RRsetInfo(DNSResponseComponent):
                 servers.sort()
             d['servers'] = servers
 
+            if map_ip_to_ns_name is not None:
+                ns_names = list(set([lb2s(map_ip_to_ns_name(s)[0][0].canonicalize().to_text()) for s in servers]))
+                ns_names.sort()
+                d['ns_names'] = ns_names
+
             tags = set()
+            nsids = set()
             for server,client in self.servers_clients:
                 for response in self.servers_clients[(server,client)]:
                     tags.add(response.effective_query_tag())
+                    nsid = response.nsid_val()
+                    if nsid is not None:
+                        nsids.add(nsid)
+
+            if nsids:
+                d['nsid_values'] = list(nsids)
+                d['nsid_values'].sort()
+
             d['query_options'] = list(tags)
             d['query_options'].sort()
 
@@ -1211,7 +1257,7 @@ class NSECSet(DNSResponseComponent):
         self.invalid_nsec3_hash = set()
         self.use_nsec3 = False
         for rrset in rrsets:
-            #XXX There shouldn't be multple NSEC(3) RRsets of the same owner
+            #XXX There shouldn't be multiple NSEC(3) RRsets of the same owner
             # name in the same response, but check for it and address it (if
             # necessary)
             assert rrset.name not in self.rrsets
