@@ -324,6 +324,10 @@ QTH_MODE_WRITE_READ = 0
 QTH_MODE_WRITE = 1
 QTH_MODE_READ = 2
 
+QTH_SETUP_DONE = 0
+QTH_SETUP_NEED_WRITE = 1
+QTH_SETUP_NEED_READ = 2
+
 class DNSQueryTransportHandler(object):
     singleton = False
     allow_loopback_query = False
@@ -361,6 +365,8 @@ class DNSQueryTransportHandler(object):
         self.start_time = None
         self.end_time = None
 
+        self.setup_state = QTH_SETUP_DONE
+
         self.qtms = []
 
     def _set_timeout(self, qtm):
@@ -531,6 +537,9 @@ class DNSQueryTransportHandler(object):
         if self._processed_queue is not None:
             self._processed_queue.put(self)
 
+    def do_setup(self):
+        pass
+
     def do_write(self):
         try:
             self.msg_send_index += self.sock.send(self.msg_send[self.msg_send_index:])
@@ -760,6 +769,11 @@ class DNSQueryTransportHandlerHTTP(DNSQueryTransportHandlerMulti):
         self.password = parse_result.password
         self.insecure = insecure
 
+        if self.use_ssl:
+            self.setup_state = QTH_SETUP_NEED_WRITE
+        else:
+            self.setup_state = QTH_SETUP_DONE
+
         af = 0
         try:
             addrinfo = socket.getaddrinfo(self.host, self.dport, af, self.transport_type)
@@ -769,18 +783,19 @@ class DNSQueryTransportHandlerHTTP(DNSQueryTransportHandlerMulti):
 
         self.chunked_encoding = None
 
-    def _create_socket(self):
-        super(DNSQueryTransportHandlerHTTP, self)._create_socket()
+    def _upgrade_socket_to_ssl(self):
+        if isinstance(self.sock.sock, ssl.SSLSocket):
+            return
 
-        if self.use_ssl:
-            #XXX this is python >= 2.7.9 only
-            ctx = ssl.create_default_context()
-            if self.insecure:
-                ctx.check_hostname = False
-                ctx.verify_mode = ssl.CERT_NONE
-            new_sock = Socket(ctx.wrap_socket(self.sock.sock, server_hostname=self.host))
-            new_sock.lock = self.sock.lock
-            self.sock = new_sock
+        #XXX this is python >= 2.7.9 only
+        ctx = ssl.create_default_context()
+        if self.insecure:
+            ctx.check_hostname = False
+            ctx.verify_mode = ssl.CERT_NONE
+        new_sock = Socket(ctx.wrap_socket(self.sock.sock, server_hostname=self.host,
+                do_handshake_on_connect=False))
+        new_sock.lock = self.sock.lock
+        self.sock = new_sock
 
     def _post_data(self):
         return 'content=' + urlquote.quote(json.dumps(self.serialize_requests()))
@@ -797,7 +812,7 @@ class DNSQueryTransportHandlerHTTP(DNSQueryTransportHandlerMulti):
 
     def init_req(self):
         data = self._post_data()
-        self.msg_send = codecs.encode('POST %s HTTP/1.1\r\nHost: %s\r\nUser-Agent: DNSViz/0.8.2\r\nAccept: application/json\r\n%sContent-Length: %d\r\nContent-Type: application/x-www-form-urlencoded\r\n\r\n%s' % (self.path, self.host, self._authentication_header(), len(data), data), 'latin1')
+        self.msg_send = codecs.encode('POST %s HTTP/1.1\r\nHost: %s\r\nUser-Agent: DNSViz/0.9.0\r\nAccept: application/json\r\n%sContent-Length: %d\r\nContent-Type: application/x-www-form-urlencoded\r\n\r\n%s' % (self.path, self.host, self._authentication_header(), len(data), data), 'latin1')
         self.msg_send_len = len(self.msg_send)
         self.msg_send_index = 0
 
@@ -806,6 +821,20 @@ class DNSQueryTransportHandlerHTTP(DNSQueryTransportHandlerMulti):
         if self.err is not None and not isinstance(self.err, SocketInUse):
             self.err = RemoteQueryTransportError('Error making HTTP connection: %s' % self.err)
 
+    def do_setup(self):
+        if self.use_ssl:
+            self._upgrade_socket_to_ssl()
+            try:
+                self.sock.sock.do_handshake()
+            except ssl.SSLWantReadError:
+                self.setup_state = QTH_SETUP_NEED_READ
+            except ssl.SSLWantWriteError:
+                self.setup_state = QTH_SETUP_NEED_WRITE
+            except ssl.SSLError as e:
+                self.err = RemoteQueryTransportError('SSL Error: %s' % e)
+            else:
+                self.setup_state = QTH_SETUP_DONE
+
     def do_write(self):
         val = super(DNSQueryTransportHandlerHTTP, self).do_write()
         if self.err is not None:
@@ -814,7 +843,11 @@ class DNSQueryTransportHandlerHTTP(DNSQueryTransportHandlerMulti):
 
     def do_read(self):
         try:
-            buf = self.sock.recv(65536)
+            try:
+                buf = self.sock.recv(65536)
+            except ssl.SSLWantReadError:
+                return False
+
             if buf == b'':
                 raise EOFError
             self.msg_recv_buf += buf
@@ -1397,7 +1430,7 @@ class _DNSQueryTransportManager:
         query_meta = {}
         expirations = []
 
-        # initalize "in" fds for select
+        # initialize "in" fds for select
         rlist_in = [self._notify_read_fd]
         wlist_in = []
         xlist_in = []
@@ -1421,11 +1454,21 @@ class _DNSQueryTransportManager:
             for fd in wlist_out:
                 qh = query_meta[fd]
 
-                if qh.do_write():
-                    if qh.err is not None or qh.mode == QTH_MODE_WRITE:
+                if qh.setup_state == QTH_SETUP_DONE:
+                    if qh.do_write():
+                        if qh.err is not None or qh.mode == QTH_MODE_WRITE:
+                            qh.cleanup()
+                            finished_fds.append(fd)
+                        else: # qh.mode == QTH_MODE_WRITE_READ
+                            wlist_in.remove(fd)
+                            rlist_in.append(qh.sock.reader_fd)
+
+                else:
+                    qh.do_setup()
+                    if qh.err is not None:
                         qh.cleanup()
                         finished_fds.append(fd)
-                    else: # qh.mode == QTH_MODE_WRITE_READ
+                    elif qh.setup_state == QTH_SETUP_NEED_READ:
                         wlist_in.remove(fd)
                         rlist_in.append(qh.sock.reader_fd)
 
@@ -1436,9 +1479,19 @@ class _DNSQueryTransportManager:
 
                 qh = query_meta[fd]
 
-                if qh.do_read(): # qh.mode in (QTH_MODE_WRITE_READ, QTH_MODE_READ)
-                    qh.cleanup()
-                    finished_fds.append(qh.sock.reader_fd)
+                if qh.setup_state == QTH_SETUP_DONE:
+                    if qh.do_read(): # qh.mode in (QTH_MODE_WRITE_READ, QTH_MODE_READ)
+                        qh.cleanup()
+                        finished_fds.append(qh.sock.reader_fd)
+
+                else:
+                    qh.do_setup()
+                    if qh.err is not None:
+                        qh.cleanup()
+                        finished_fds.append(fd)
+                    elif qh.setup_state in (QTH_SETUP_NEED_WRITE, QTH_SETUP_DONE):
+                        rlist_in.remove(fd)
+                        wlist_in.append(qh.sock.writer_fd)
 
             # handle the expired queries
             future_index = bisect.bisect_right(expirations, ((time.time(), DNSQueryTransportHandlerWrapper(None))))
@@ -1498,12 +1551,17 @@ class _DNSQueryTransportManager:
                             query_meta[qh.sock.reader_fd] = qh
                             query_meta[qh.sock.writer_fd] = qh
                             bisect.insort(expirations, (qh.expiration, DNSQueryTransportHandlerWrapper(qh)))
-                            if qh.mode in (QTH_MODE_WRITE_READ, QTH_MODE_WRITE):
+                            if qh.setup_state == QTH_SETUP_DONE:
+                                if qh.mode in (QTH_MODE_WRITE_READ, QTH_MODE_WRITE):
+                                    wlist_in.append(qh.sock.writer_fd)
+                                elif qh.mode == QTH_MODE_READ:
+                                    rlist_in.append(qh.sock.reader_fd)
+                                else:
+                                    raise Exception('Unexpected mode: %d' % qh.mode)
+                            elif qh.setup_state == QTH_SETUP_NEED_WRITE:
                                 wlist_in.append(qh.sock.writer_fd)
-                            elif qh.mode == QTH_MODE_READ:
+                            elif qh.setup_state == QTH_SETUP_NEED_READ:
                                 rlist_in.append(qh.sock.reader_fd)
-                            else:
-                                raise Exception('Unexpected mode: %d' % qh.mode)
                     except queue.Empty:
                         break
 
